
// Arduino interface code for ARDOP running on a Teensy 3.6

#include "LinkitConfig.h"
#include "ConfigARDOP.h"

#define ARDOP
#define HOSTPORT Serial
#define MONPORT Serial

#define LED1 LED_BUILTIN
#define LED2 LED_BUILTIN
#define LED3 LED_BUILTIN
#define LED4 LED_BUILTIN

#define CPU_RESTART_ADDR (uint32_t *)0xE000ED0C
#define CPU_RESTART_VAL 0x5FA0004
#define CPU_RESTART (*CPU_RESTART_ADDR = CPU_RESTART_VAL);

unsigned char RXBUFFER[500];	// Async RX Buffer. Enough for Stuffed Host Mode Frame

extern volatile int RXBPtr;
volatile int flag = 0;
volatile int flag2 = 0;
volatile int lastperiod = 0, lastint = 0;

int SerialHost = TRUE;				// Will eventually allow switching from Serial to I2c without reconfig
int ActivePort = 0;						// Serial port in use
int SerialWatchDog = 0;
extern "C" int blnClosing;

int loadCounter = 0;		// for performance monitor
int lastLoadTicks = 0;
unsigned int lastRTCTick = 0;	// Make sure this is set to ticks mod 1000 so RTC incrememtns at mS 0


int VRef;
extern "C" int TXLevel;

void CommonSetup();
void setupPDB(int SampleRate);
void setupDAC();
void setupADC(int Pin);
extern "C" void StartDAC();
extern "C" void StopDAC();
void StartADC();
void setupOLED();
void setupKMR_18();

void setupWDTTFT();

// Arduino is a c++ environment so functions in ardop must be defined as "C"

extern "C"
{
  void WriteDebugLog(int Level, const char * format, ...);
  void InitSound();
  void HostInit();
  void CheckTimers();
  void HostPoll();
  void MainPoll();
  void InitDMA();
  void PlatformSleep();
  void AdjustTXLevel(int Level);
  void KISSInit();
  void ProcessBlock();

  void ProcessSCSPacket(unsigned char * rxbuffer, int Length);

#include "C:\OneDrive\Dev\Source\TeensyProjects\ARDOPC/ARDOPC.h"

  extern unsigned int tmrPollOBQueue;

  extern int UseKISS;

#define Now getTicks()

#define MEM_LEN 512
  extern uint8_t databuf[MEM_LEN];
  extern volatile int i2cputptr, i2cgetptr;

  void HostPoll()
  {
#ifdef I2CHOST

    while (i2cgetptr != i2cputptr)
    {
      unsigned char c;
      c = databuf[i2cgetptr++];
      i2cgetptr &= 0x1ff;				// 512 cyclic

      RXBUFFER[RXBPtr++] = c;

      if (i2cgetptr == i2cputptr || RXBPtr > 498)
      {
        ProcessSCSPacket(RXBUFFER, RXBPtr);
        return;
      }
    }
#else

#ifdef i2cSlaveSupport
    i2cloop();					// I2C but not for Host
#endif

    // To simplify switching USB / Serial / Bluetooth / ESP we poll all defined ports,
    // and switch to using whichever provides input

    // Both boards could use USB(Serail) or Serial1
    // PIBoard has ESP on Serail3
    // WDT Board has BT on Serial5

    int Avail = HOSTPORT.available();
    int Count;

    if (Avail)
    {
      if (Avail > (499 - RXBPtr))
        Avail = 499 - RXBPtr;

      Count = HOSTPORT.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
      RXBPtr += Count;
      if (ActivePort != 1)
      {
        ActivePort = 1;
        WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
      }

      ProcessSCSPacket(RXBUFFER, RXBPtr);
    }
#ifdef HOSTPORT2
    else
    {
      Avail = HOSTPORT2.available();
      if (Avail)
      {
        if (Avail > (499 - RXBPtr))
          Avail = 499 - RXBPtr;

        Count = HOSTPORT2.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
        RXBPtr += Count;
        if (ActivePort != 2)
        {
          ActivePort = 2;
          WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
        }
        ProcessSCSPacket(RXBUFFER, RXBPtr);
      }
    }
#endif
#ifdef HOSTPORT3
    else
    {
      Avail = HOSTPORT3.available();
      if (Avail)
      {
        if (Avail > (499 - RXBPtr))
          Avail = 499 - RXBPtr;

        Count = HOSTPORT3.readBytes((char *)&RXBUFFER[RXBPtr], Avail);
        RXBPtr += Count;
        if (ActivePort != 3)
        {
          ActivePort = 3;
          WriteDebugLog(LOGINFO, "Input is on Port %d", ActivePort);
        }
        ProcessSCSPacket(RXBUFFER, RXBPtr);
      }
    }
#endif
#endif		// I2CHOST
  }
}

void Sleep(int mS)
{
#if defined MONPORT && defined CPULOAD

  int loadtime = millis() - lastLoadTicks;

  loadCounter += mS;

  if (loadtime > 999)
  {
    MONprintf("Load = %d" , 100 - (100 * loadCounter / loadtime));
    lastLoadTicks = millis();
    loadCounter = 0;
  }
#endif
  delay(mS);
}

void setup()
{
  uint32_t i, sum = 0;

#ifdef I2CHOST
  SerialHost = FALSE;
#endif

  //  CommonSetup();

  HOSTPORT.begin(115200);
  Serial1.begin(115200);
  

  WriteDebugLog(LOGALERT, "ARDOPC Version %s CPU %d Bus %d FreeRAM %d", ProductVersion, 0, 0, FreeRam());
  blnTimeoutTriggered = FALSE;
  SetARDOPProtocolState(DISC);

  //  InitSound();
  HostInit();
  tmrPollOBQueue = Now + 10000;
  ProtocolMode = ARQ;

  //  if (UseKISS)
  //			KISSInit();

  // Configure the ADC and run at least one software-triggered
  // conversion.  This completes the self calibration stuff and
  // leaves the ADC in a state that's mostly ready to use

  //  analogReadRes(16);
  //analogReference(INTERNAL); // range 0 to 1.2 volts
  //  analogReference(DEFAULT); // range 0 to 3.3 volts
  //analogReadAveraging(8);
  // Actually, do many normal reads, to start with a nice DC level

  for (i = 0; i < 1024; i++)
  {
    sum += analogRead(16);
  }

  WriteDebugLog(LOGDEBUG, "DAC Baseline %d", sum / 1024);

  // setupPDB(12000);			// 12K sample rate
  // setupDAC();
  // setupADC(16);
  // StartADC();

#ifdef PIBOARD

  // PI Baord can read Analog out which allows u to Read Vref

  SetPot(1, 256);				// TX Level Gain = 1

  delay(100);

  analogRead(17);

  for (i = 0; i < 100; i++)
  {
    VRef += analogRead(17);
  }
  VRef /= 100;
  analogRead(16);		// Set ADC back to A0

#endif


  WriteDebugLog(0, "VREF %d offset %d", VRef, VRef - 32768);

  // AdjustTXLevel(TXLevel);

  print_mac();

#ifdef PLOTCONSTELLATION
#ifdef OLED
  setupOLED();
#endif
#ifdef WDTTFT
  setupWDTTFT();
#endif
#ifdef KMR_18
  setupKMR_18();
#endif

#endif
}

int lastticks = Now;

void loop()
{
  //  PollReceivedSamples();
  //  CheckTimers();
  // HostPoll();
  // MainPoll();

  if (Now - lastticks > 999)
  {
    // Debug 1 sec tick
  }
  //PlatformSleep();
  //RadioPoll();
//    ProcessBlock();
  Sleep(1);

#ifdef I2CHOST
  SerialWatchDog++;		// Reset if nothing for 60 secs
#endif

  if (blnClosing || SerialWatchDog > 60000)
  {
    CPU_RESTART // reset processor
  }
}

extern "C"
{
  int OKtoAdjustLevel()
  {
    // Only auto adjust level when disconnected.
    // Level is set at end of each received packet when connected

    return (ProtocolState == DISC);
  }

  void StartCapture()
  {
    Capturing = TRUE;
    DiscardOldSamples();
    ClearAllMixedSamples();
    State = SearchingForLeader;
  }
  void StopCapture()
  {
    Capturing = FALSE;
  }
  void TurnroundLink()
  {
    if (blnEnbARQRpt > 0 || blnDISCRepeating)	// Start Repeat Timer if frame should be repeated
      dttNextPlay = Now + intFrameRepeatInterval;
  }
}



// function from the sdFat library (SdFatUtil.cpp)
// licensed under GPL v3
// Full credit goes to William Greiman.

extern "C" char* sbrk(int incr);

char *__brkval = 0;
char __bss_end = 0;


int FreeRam()
{
  char top;

  return __brkval ? &top - __brkval : &top - &__bss_end;
}


// Function to get mac/serial number

uint8_t mac[8];

// http://forum.pjrc.com/threads/91-teensy-3-MAC-address

void readserialno(uint8_t *mac)
{
  /*
    noInterrupts();

    // With 3.6 have to read a block of 8 bytes at zero
    // With 3.1 two blocks of 4 at e and f

    FTFL_FCCOB0 = 0x41;             // Selects the READONCE command
    #if defined(__MK64FX512__) || defined(__MK66FX1M0__)
    FTFL_FCCOB1 = 0;             // read the given word of read once area
    #else
    FTFL_FCCOB1 = 0x0F;          // read the given word of read once area
    #endif
    FTFL_FCCOB2 = 0;
    FTFL_FCCOB3 = 0;
    // launch command and wait until complete
    FTFL_FSTAT = FTFL_FSTAT_CCIF;
    while (!(FTFL_FSTAT & FTFL_FSTAT_CCIF));

    #if defined(__MK64FX512__) || defined(__MK66FX1M0__)

    // Top 3 bytes (Manufacture ID in bits 8-31

     (mac++) = FTFL_FCCOB5;
     (mac++) = FTFL_FCCOB6;
     (mac++) = FTFL_FCCOB7;

  	// Serial No in bits 8-31

     (mac++) = FTFL_FCCOB9;
     (mac++) = FTFL_FCCOBA;
     (mac++) = FTFL_FCCOBB;
    #else

    // Could read Manufacture ID from page E but no point as fixed (04:E9:E5)

     (mac++) = 0x04;
     (mac++) = 0xE9;
     (mac++) = 0xE5;
     (mac++) = FTFL_FCCOB5;
     (mac++) = FTFL_FCCOB6;
     (mac++) = FTFL_FCCOB7;


    #endif

    interrupts();
  */
}

void print_mac()
{
  readserialno(mac);
  WriteDebugLog(6, "Hardware Serial No %02X:%02X:%02X:%02X:%02X:%02X",
                mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
}

#include <stdio.h>
#include <stdarg.h>

void WriteDebugLog(int Level, const char * format, ...)
{
  // Use Dragon log format (stats with 32 bit timestamp)

  char Mess[256];
  va_list(arglist);
  int len;

  va_start(arglist, format);
#ifdef LOGTOHOST

  // correct RTC if needed

  while ((millis() - lastRTCTick) > 999)
  {
    lastRTCTick += 1000;
    RTC++;
  }
  Mess[0] = (RTC >> 24) & 0xff;
  Mess[1] = (RTC >> 16) & 0xff;
  Mess[2] = (RTC >> 8) & 0xff;
  Mess[3] = RTC & 0xff;

  sprintf(&Mess[4], "%03d", millis() % 1000);	// millisecs
  Mess[7] = '0' + Level;

  len = vsnprintf(&Mess[8], sizeof(Mess) - 8, format, arglist);
  strcat(&Mess[8], "\r\n");
  SendLogToHost(Mess, len + 10);
#endif
#ifdef MONPORT
  vsprintf(Mess, format, arglist);
  MONPORT.println(Mess);
#endif
#ifdef UIMON
  vsnprintf(Mess, sizeof(Mess), format, arglist);
  SendMonUI(Mess);
#endif
  return;
}

// Write to Log, either via host or serial port
void MONprintf(const char * format, ...)
{
  char Mess[256];
  va_list(arglist);
  int len;

  va_start(arglist, format);
#ifdef LOGTOHOST

  // correct RTC if needed

  while ((millis() - lastRTCTick) > 999)
  {
    lastRTCTick += 1000;
    RTC++;
  }
  Mess[0] = RTC >> 24;
  Mess[1] = RTC >> 16;
  Mess[2] = RTC >> 0;
  Mess[3] = RTC;

  sprintf(&Mess[4], "%03d", millis() % 1000);	// millisecs
  Mess[7] = '0';
  len = vsprintf(&Mess[8], format, arglist);
  strcat(&Mess[8], "\r\n");
  SendLogToHost(Mess, len + 10);
#endif
#ifdef MONPORT
  vsprintf(Mess, format, arglist);
  MONPORT.println(Mess);
#endif
#ifdef UIMON
  vsnprintf(Mess, sizeof(Mess), format, arglist);
  SendMonUI(Mess);
#endif    return;
}

void CloseDebugLog()
{
}

void CloseStatsLog()
{
}

extern "C" void HostInit()
{
  return;
}


unsigned int getTicks()
{
  return millis();
}

extern "C" void SetLED(int LED, int Val)
{}

extern "C" void displayState(const char * State)
{}

extern "C" void PutString(const char * Msg)
{
  SerialSendData((const uint8_t *)Msg, strlen(Msg));
}

extern "C" void SerialSendData(const uint8_t * Msg, int Len)
{
  HOSTPORT.write(Msg, Len);
}


